from __future__ import unicode_literals

import datetime

from django.contrib.gis.db import models
from django.utils.translation import ugettext_lazy as _

from . import attrs


class Unit(models.Model):
    """
    A unit for a given attribute
    """
    name = models.CharField(_('name'), max_length=100, db_index=True)
    symbol = models.CharField(_('symbol'), max_length=10)

    def __unicode__(self):
        return self.name


class Attribute(models.Model):
    """
    An attribute for a model (the 'A' in EAV/OAV)
    """
    TYPE_TEXT = 1
    TYPE_BOOLEAN = 2
    TYPE_INTEGER = 3
    TYPE_DECIMAL = 4
    TYPE_DATE = 5
    TYPE_TIME = 6

    BOOLEAN_TRUE_TEXTS = ('TRUE', 'YES', 'T', 'Y', '1',)
    BOOLEAN_FALSE_TEXTS = ('FALSE', 'NO', 'F', 'N', '0',)
    BOOLEAN_NULL_TEXTS = ('NULL', '',)

    CHOICES_FOR_TYPE = (
        (TYPE_TEXT, _('text')),
        (TYPE_BOOLEAN, _('boolean')),
        (TYPE_INTEGER, _('integer')),
        (TYPE_DECIMAL, _('decimal')),
        (TYPE_DATE, _('date')),
        (TYPE_TIME, _('time')),
    )

    slug = models.SlugField(verbose_name=_('slug'), unique=True)
    name = models.CharField(_('name'), max_length=100, db_index=True)
    type = models.PositiveSmallIntegerField(_('type'), choices=CHOICES_FOR_TYPE)
    unit = models.ForeignKey('Unit', verbose_name=_('unit'), null=True, blank=True)

    def get_name_display(self):
        if self.name:
            return self.name
        return self.slug

    def get_label(self):
        label_parts = [self.get_name_display(), ]
        if self.unit:
            label_parts.append('({symbol})'.format(symbol=self.unit.symbol))
        return ' '.join(label_parts)

    def get_choices(self):
        if self.type == self.TYPE_BOOLEAN:
            return (
                ('', _('unknown')),
                ('TRUE', _('yes')),
                ('FALSE', _('no')),
            )
        return [(choice.pk, choice.get_value_display()) for choice in self.choice_set.order_by('sort_order', 'name')]

    def text_to_value(self, text):
        if self.type == self.TYPE_TEXT:
            return text
        t = text.upper().strip()
        if self.type == self.TYPE_INTEGER:
            return int(t)
        if self.type == self.TYPE_DECIMAL:
            return float(t)
        if self.type == self.TYPE_BOOLEAN:
            if t in self.BOOLEAN_TRUE_TEXTS:
                return True
            if t in self.BOOLEAN_FALSE_TEXTS:
                return False
            if t in self.BOOLEAN_NULL_TEXTS:
                return None
        if self.type == self.TYPE_DATE:
            parts = text.split('-')
            return datetime.date(
                int(parts[0].strip()),
                int(parts[1].strip()),
                int(parts[2].strip()),
            )
        if self.type == self.TYPE_TIME:
            parts = text.split(':')
            hours = int(parts[0].strip())
            try:
                minutes = int(parts[1].strip())
            except IndexError:
                minutes = 0
            try:
                seconds = int(parts[2].strip())
            except IndexError:
                seconds = 0
            return datetime.time(hours, minutes, seconds)
        # We cannot parse this
        raise ValueError('Cannot convert text "{text}" to value of type {type}.'.format(
            text=text,
            type=self.get_type_display()
        ))

    def __unicode__(self):
        return self.get_name_display()


class Choice(models.Model):
    attribute = models.ForeignKey(Attribute)
    value = models.CharField(_('value'), max_length=100, blank=True)
    name = models.CharField(_('name'), max_length=100, blank=True)
    description = models.TextField(_('description'), blank=True)
    sort_order = models.IntegerField(_('sort order'), default=0)

    def get_value_display(self):
        if self.name:
            return self.name
        return self.value

    def __unicode__(self):
        return '{attribute}.{name}'.format(attribute=self.attribute.name, name=self.get_value_display())

    class Meta:
        ordering = ['attribute_id', 'sort_order', 'value', 'pk', ]


class AbstractModelAttribute(models.Model):
    attribute = models.ForeignKey(Attribute)
    value = models.CharField(_('value'), max_length=100, blank=True)

    def get_value_display(self):
        value = '{}'.format(self.value)
        if self.attribute.choice_set.exists():
            choice_values = []
            for choice in value.split(','):
                try:
                    choice_id = int(choice)
                except ValueError:
                    choice_values.append('!!!error!!!')
                else:
                    try:
                        choice = Choice.objects.get(pk=choice_id)
                    except Choice.DoesNotExist:
                        choice_values.append('!!!not_found!!!')
                    else:
                        choice_values.append(choice.get_value_display())
            value = ', '.join(choice_values)
        return value

    def __unicode__(self):
        return '{attribute} = {value}'.format(
            attribute=self.attribute,
            value=self.get_value_display(),
        )

    class Meta:
        abstract = True


def create_model_attribute_class(model, class_name=None, related_name=None, meta=None):
    """
    Generate a value class (derived from AbstractModelAttribute) for a given model class
    :param model: The model to create a AbstractModelAttribute class for
    :param class_name: The name of the AbstractModelAttribute class to generate
    :param related_name: The related name
    :return: A model derives from AbstractModelAttribute with an object pointing to model_class
    """

    if model._meta.abstract:
        # This can't be done, because `object = ForeignKey(model_class)` would fail.
        raise TypeError("Can't create mav for abstract class {0}".format(model.__name__))

    # Define inner Meta class
    if not meta:
        meta = {}

    # Force only one value for each model, attribute set
    meta['unique_together'] = list(meta.get('unique_together', [])) + [('attribute', 'object')]

    # Use the same tablespace as the model
    meta['db_tablespace'] = model._meta.db_tablespace

    # Use same setting for managed as model
    meta['managed'] = model._meta.managed

    # Set the default db table
    meta.setdefault('db_table', '{0}_attr'.format(model._meta.db_table))

    # The name of the class to generate
    if class_name is None:
        value_class_name = '{name}Attr'.format(name=model.__name__)
    else:
        value_class_name = class_name

    # The related name to set
    if related_name is None:
        model_class_related_name = 'attrs'
    else:
        model_class_related_name = related_name

    # Make a type for our class
    value_class = type(
        str(value_class_name),
        (AbstractModelAttribute,),
        dict(
            # Set to same module as model_class
            __module__=model.__module__,
            # Add a foreign key to model_class
            object=models.ForeignKey(
                model,
                related_name=model_class_related_name
            ),
            # Add Meta class
            Meta=type(
                str('Meta'),
                (object,),
                meta
            ),
        ))

    return value_class


def add_mav_to(model, class_name=None, related_name=None, meta=None):
    """
    Patch model class to have mav attributes
    :param model: The model class to patch
    :param class_name: The name of the class to generate
    :param related_name: The related_name to set in the model
    :return: The generated class
    """

    # Generate the class
    mav_class = create_model_attribute_class(
        model=model,
        class_name=class_name,
        related_name=related_name,
        meta=meta,
    )

    # Add it to .attrs
    setattr(attrs, mav_class.__name__, mav_class)

    # Link it to the model
    model._mav_class = mav_class

    # Return the generated class
    return mav_class



# coding: utf-8
from __future__ import unicode_literals

import datetime
from django.db import models

from django.test import TestCase

from model_mommy import mommy
from mav.decorators import mav

from .models import Attribute

@mav
class Foo(models.Model):
    name = models.CharField(max_length=100)


class ValueTestCase(TestCase):
    def test_unit_symbol(self):
        """
        Test display of strange symbols in units
        """
        unit_celsius = mommy.make('mav.Unit', name='degrees Celsius', symbol='° C')
        self.assertEqual('° C', unit_celsius.symbol)

    def test_attribute_type_text(self):
        """
        Test parsing of text to text value
        """
        attribute = mommy.make(Attribute, type=Attribute.TYPE_TEXT)
        self.assertEqual('text value', attribute.text_to_value('text value'))

    def test_attribute_type_integer(self):
        """
        Test parsing of text to integer value
        """
        attribute = mommy.make(Attribute, type=Attribute.TYPE_INTEGER)
        self.assertEqual(189, attribute.text_to_value('189'))
        with self.assertRaises(ValueError):
            attribute.text_to_value('189.1')
        with self.assertRaises(ValueError):
            attribute.text_to_value('this.is.not.an.integer')

    def test_attribute_type_decimal(self):
        """
        Test parsing of text to decimal value
        """
        attribute = mommy.make(Attribute, type=Attribute.TYPE_DECIMAL)
        self.assertEqual(189, attribute.text_to_value('189'))
        self.assertEqual(189.1, attribute.text_to_value('189.1'))
        with self.assertRaises(ValueError):
            attribute.text_to_value('this.is.not.a.decimal')

    def test_attribute_type_boolean(self):
        """
        Test parsing of text to boolean value
        """
        attribute = mommy.make(Attribute, type=Attribute.TYPE_BOOLEAN)
        for text in Attribute.BOOLEAN_TRUE_TEXTS:
            self.assertEqual(True, attribute.text_to_value(text))
        for text in Attribute.BOOLEAN_FALSE_TEXTS:
            self.assertEqual(False, attribute.text_to_value(text))
        for text in Attribute.BOOLEAN_NULL_TEXTS:
            self.assertEqual(None, attribute.text_to_value(text))
        with self.assertRaises(ValueError):
            attribute.text_to_value('this.is.not.a.boolean')

    def test_attribute_type_date(self):
        """
        Test parsing of text to date value
        """
        date = datetime.date(2000, 2, 29)
        date_texts = ['2000-02-29', '2000-2-29', ]
        attribute = mommy.make(Attribute, type=Attribute.TYPE_DATE)
        for text in date_texts:
            self.assertEqual(date, attribute.text_to_value(text))
        with self.assertRaises(ValueError):
            attribute.text_to_value('this.is.not.a.date')

    def test_attribute_type_time(self):
        """
        Test parsing of text to time value
        """
        time = datetime.time(23, 1, 15)
        time_texts = ['23:01:15', '23:1:15', ]
        attribute = mommy.make(Attribute, type=Attribute.TYPE_TIME)
        for text in time_texts:
            self.assertEqual(time, attribute.text_to_value(text))
        with self.assertRaises(ValueError):
            attribute.text_to_value('this.is.not.a.time')

    def test_class_with_attrs(self):
        """
        Test to see if class with attrs behaves
        """
        bar = Attribute(type=Attribute.TYPE_TEXT, slug='bar')
        bar.save()
        foo = Foo(name='foo')
        foo.save()
        self.assertEqual(bar.slug, 'bar')
        mav_class = Foo._mav_class
        from mav.attrs import FooAttr
        self.assertEqual(FooAttr, mav_class)
        attr = mav_class(attribute=bar, object=foo, value='foobar')
        attr.save()
        self.assertEqual(attr.value, 'foobar')
        self.assertEqual(attr.object, foo)
        self.assertEqual(attr.attribute, bar)



from __future__ import unicode_literals

from django import forms
from django.contrib import admin

from .models import Attribute, Choice, Unit


try:
    from modeltranslation.admin import TranslationAdmin as AdminClass
except ImportError:
    AdminClass = admin.ModelAdmin


admin.site.register(Attribute, AdminClass)
admin.site.register(Choice, AdminClass)
admin.site.register(Unit, AdminClass)



from django.utils.text import slugify


def slugify_with_underscores(text):
    """
    A slugify that does not tolerate dashes (replaced with underscores)
    :param text: The text to slugify
    :return: The slugified text
    """
    return slugify(text).replace('-', '_')



from __future__ import unicode_literals, absolute_import

from .models import add_mav_to


def _mav(cls, *args, **kwargs):
    """
    Add mav to a model class and return that model class
    :param cls: The model class that needs mav
    :return: The model class
    """
    add_mav_to(cls, *args, **kwargs)
    return cls


def mav(cls=None, **kwargs):
    """
    Smart decorator class to add mav to a model class, and can optionally take arguments
    :param cls: The model class that needs mav
    :return: Class (when called with cls argument) or wrapper class (when called without cls argument)
    """
    if cls:
        return _mav(cls)
    else:
        def wrapper(cls):
            return _mav(cls, **kwargs)
        return wrapper



from __future__ import unicode_literals
from collections import OrderedDict

from django import forms
from django.template.defaultfilters import capfirst

from .models import Attribute, Choice, AbstractModelAttribute


FIELD_PREFIX = '__mav__'


class RelaxedFloatField(forms.FloatField):
    """
    Allow both comma and point as decimal separator, unless you use both
    """

    def clean(self, value):
        value = unicode(value)
        if ',' in value and not '.' in value:
            value = value.replace(',', '.')
        return super(RelaxedFloatField, self).clean(value)


def generate_attribute_field(attribute, value=''):
    """
    Generate a attribute field
    """
    choices = attribute.get_choices()
    if choices:
        required = True
        for choice in choices:
            if choice[0] == '':
                required = False
        field = forms.ChoiceField(choices=choices, required=required, initial=value)
    elif attribute.type == Attribute.TYPE_INTEGER:
        field = forms.IntegerField(required=False, initial=value)
    elif attribute.type == Attribute.TYPE_DECIMAL:
        field = RelaxedFloatField(required=False, initial=value)
    elif attribute.type == Attribute.TYPE_DATE:
        field = forms.DateField(required=False, initial=value)
        field.widget.attrs['data-datepicker'] = True
    else:
        field = forms.CharField(required=False, initial=value)
    field.label = capfirst(attribute.get_label())
    return field


def add_attribute_field_to_form(form, attribute, value=''):
    """
    Add field for attribute to form
    """
    field_name = '{prefix}{id}'.format(prefix=FIELD_PREFIX, id=attribute.id)
    field = generate_attribute_field(attribute, value)
    form.fields[field_name] = field
    return form


def add_attribute_fields_to_form(form, attributes, attrs):
    """
    Add fields for attributes to form
    form = form to modify
    attributes = attributes to add (Attribute)
    attrs = existing attributes (AbstractModelAttribute)
    """
    existing = OrderedDict()
    # Find existing attributes
    if attrs:
        for attr in attrs:
            attribute = attr.attribute
            existing[attribute.id] = attr.value
    # Add fields for the attributes
    if attributes:
        for attribute in attributes:
            value = existing.pop(attribute.id, '')
            add_attribute_field_to_form(form, attribute, value)
    # See if we have attributes for unlisted attributes and add those too
    for attribute_id in existing:
        value = existing[attribute_id]
        attribute = Attribute.objects.get(pk=attribute_id)
        add_attribute_field_to_form(form, attribute, value)
    # Return the form
    return form


def save_attribute_fields(form):
    """
    Save the attribute fields of a form to attributes
    """
    instance = form.instance
    attributes = {}
    pos = len(FIELD_PREFIX)
    for k in form.cleaned_data:
        if k[0:pos] == FIELD_PREFIX:
            attribute_id = k[pos:]
            attributes[attribute_id] = form.cleaned_data[k]
    if attributes:
        # Process all attributes
        for attribute_id in attributes:
            # We need content_type, see above
            value, created = instance.attrs.get_or_create(
                attribute_id=attribute_id,
                object=instance,
            )
            # Set the value (don't allow None)
            value.value = attributes[attribute_id]
            if value.value is None:
                value.value = ''
            # And save the value
            value.save()


class AttrsModelFormMixin(object):
    """
    Mixin to handle attributes in a ModelForm
    """

    def __init__(self, *args, **kwargs):
        # Create the form
        super(AttrsModelFormMixin, self).__init__(*args, **kwargs)
        # Modify the form
        add_attribute_fields_to_form(self, self.get_attributes(), self.get_attrs())

    def get_attributes(self):
        """
        Get all attributes that should be on the form
        """
        try:
            return self.instance.get_attributes()
        except AttributeError:
            return None

    def get_attrs(self):
        """
        Get existing attributes with values
        """
        try:
            return self.instance.attrs.all()
        except AttributeError:
            return None

    def save(self, commit=True):
        instance = super(AttrsModelFormMixin, self).save(commit=commit)
        if commit:
            save_attribute_fields(self)
        else:
            # Modify the save_m2m of the model
            super_save_m2m = self.save_m2m
            def save_m2m():
                super_save_m2m()
                save_attribute_fields(self)
            self.save_m2m = save_m2m
        return instance


class ModelFormWithAttrs(AttrsModelFormMixin, forms.ModelForm):
    """
    A ModelForm that also handles attributes and attributes
    """
    pass


#!/usr/local/bin/python
# -*- coding: utf-8 -*-

import os
import sys

import mav

from setuptools import setup

# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# Read version from app
version = mav.__version__

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

if sys.argv[-1] == 'publish':
    os.system('cd docs && make html')
    os.system('python setup.py sdist upload')
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()

if sys.argv[-1] == 'test':
    os.system('pip install -U -r requirements.txt')
    os.system('python manage.py test')
    sys.exit()

setup(
    name='django-mav',
    version=version,
    description="""Charts for Django made simple""",
    long_description=readme + '\n\n' + history,
    author='Dylan Verheul',
    author_email='dylan@zostera.nl',
    url='https://github.com/zostera/django-charts',
    packages=[
        "mav",
    ],
    include_package_data=True,
    install_requires=[
        "Django > 1.4",
    ],
    license="MIT License",
    zip_safe=False,
    keywords='django-mav',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'Environment :: Web Environment',
        'Framework :: Django',
    ],
)
